@@ -1,12 +1,14 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
 # Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2012-2014 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2012-2016 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import sys, os
+import sys
 import logging
+from xpra.util import envbool
+from xpra.os_util import OSX, WIN32
 from xpra.log import Logger, CaptureHandler
 log = Logger("opengl")
 
@@ -14,80 +16,109 @@ required_extensions = ["GL_ARB_texture_rectangle", "GL_ARB_vertex_program"]
 
 
 WHITELIST = {}
-BLACKLIST = {"vendor" : ["nouveau", "Humper", "VMware, Inc."],
-             "renderer" : ["Software Rasterizer"]}
+GREYLIST = {
+            "vendor"    : ["Intel", "VMware, Inc.", "Humper"]
+            }
+VERSION_REQ = {
+               "nouveau" : [3, 0],      #older versions have issues
+               }
+BLACKLIST = {
+             "renderer" : ["Software Rasterizer"]
+             }
+
+from xpra.os_util import getUbuntuVersion
+uv = getUbuntuVersion()
+if uv and uv<[15]:
+    #Ubuntu 14.x drivers are just too old
+    GREYLIST.setdefault("vendor", []).append("X.Org")
 if False:
     #for testing:
-    BLACKLIST["vendor"].append("NVIDIA Corporation")
+    GREYLIST["vendor"].append("NVIDIA Corporation")
     WHITELIST["renderer"] = ["GeForce GTX 760/PCIe/SSE2"]
-    
-if sys.platform.startswith("darwin") or sys.platform.startswith("win"):
-    #crashes were reported with the Intel driver on OSX 
-    BLACKLIST["vendor"].append("Intel Inc.")
+
+    if OSX:
+        #frequent crashes on osx with GT 650M: (see ticket #808)
+        GREYLIST.setdefault("vendor", []).append("NVIDIA Corporation")
 
 
-DEFAULT_ALPHA = not sys.platform.startswith("win") and not sys.platform.startswith("darwin")
-GL_ALPHA_SUPPORTED = os.environ.get("XPRA_ALPHA", DEFAULT_ALPHA) in (True, "1")
-DEFAULT_DOUBLE_BUFFERED = 0
-if sys.platform.startswith("win"):
-    #needed on win32?
-    DEFAULT_DOUBLE_BUFFERED = 1
-DOUBLE_BUFFERED = os.environ.get("XPRA_OPENGL_DOUBLE_BUFFERED", str(DEFAULT_DOUBLE_BUFFERED))=="1"
+#alpha requires gtk3 or *nix only for gtk2:
+DEFAULT_ALPHA = sys.version>'3' or (not WIN32 and not OSX)
+GL_ALPHA_SUPPORTED = envbool("XPRA_ALPHA", DEFAULT_ALPHA)
+#not working with gtk3 yet?
+CAN_DOUBLE_BUFFER = True
+#needed on win32?:
+DEFAULT_DOUBLE_BUFFERED = WIN32 and CAN_DOUBLE_BUFFER
+DOUBLE_BUFFERED = envbool("XPRA_OPENGL_DOUBLE_BUFFERED", DEFAULT_DOUBLE_BUFFERED)
+
+from xpra.gtk_common.gtk_util import STATIC_GRAY, GRAYSCALE, STATIC_COLOR, PSEUDO_COLOR, TRUE_COLOR, DIRECT_COLOR
+VISUAL_NAMES = {
+                STATIC_GRAY      : "STATIC_GRAY",
+                GRAYSCALE        : "GRAYSCALE",
+                STATIC_COLOR     : "STATIC_COLOR",
+                PSEUDO_COLOR     : "PSEUDO_COLOR",
+                TRUE_COLOR       : "TRUE_COLOR",
+                DIRECT_COLOR     : "DIRECT_COLOR",
+                }
+
+from xpra.gtk_common.gtk_util import LSB_FIRST, MSB_FIRST
+
+VISUAL_TYPES = {
+                LSB_FIRST   : "LSB",
+                MSB_FIRST   : "MSB",
+                }
+
+from xpra.client.gl.gtk_compat import MODE_RGBA, MODE_ALPHA, MODE_RGB, MODE_DOUBLE, MODE_SINGLE, MODE_DEPTH
+
 
 def get_visual_name(visual):
-    import gtk.gdk
     if not visual:
         return ""
-    return {
-           gtk.gdk.VISUAL_STATIC_GRAY   : "STATIC_GRAY",
-           gtk.gdk.VISUAL_GRAYSCALE     : "GRAYSCALE",
-           gtk.gdk.VISUAL_STATIC_COLOR  : "STATIC_COLOR",
-           gtk.gdk.VISUAL_PSEUDO_COLOR  : "PSEUDO_COLOR",
-           gtk.gdk.VISUAL_TRUE_COLOR    : "TRUE_COLOR",
-           gtk.gdk.VISUAL_DIRECT_COLOR  : "DIRECT_COLOR"}.get(visual.type, "unknown")
+    global VISUAL_NAMES
+    return VISUAL_NAMES.get(visual.type, "unknown")
 
 def get_visual_byte_order(visual):
-    import gtk.gdk
     if not visual:
         return ""
-    return {
-            gtk.gdk.LSB_FIRST   : "LSB",
-            gtk.gdk.MSB_FIRST   : "MSB"}.get(visual.byte_order, "unknown")
+    global VISUAL_TYPES
+    return VISUAL_TYPES.get(visual.byte_order, "unknown")
 
 def visual_to_str(visual):
     if not visual:
         return ""
-    d = {"type"         : get_visual_name(visual),
-         "byte_order"   : get_visual_byte_order(visual)}
+    d = {
+         "type"         : get_visual_name(visual),
+         "byte_order"   : get_visual_byte_order(visual),
+         }
     for k in ("bits_per_rgb", "depth"):
         d[k] = getattr(visual, k)
     return str(d)
 
 def get_DISPLAY_MODE(want_alpha=GL_ALPHA_SUPPORTED):
-    import gtk.gdkgl
-    #gtk.gdkgl.MODE_DEPTH
-    mode = 0
+    #MODE_DEPTH
     if want_alpha:
-        mode = mode | gtk.gdkgl.MODE_RGBA | gtk.gdkgl.MODE_ALPHA
+        mode = MODE_RGBA | MODE_ALPHA
     else:
-        mode = mode | gtk.gdkgl.MODE_RGB
+        mode = MODE_RGB
     if DOUBLE_BUFFERED:
-        mode = mode | gtk.gdkgl.MODE_DOUBLE
+        mode = mode | MODE_DOUBLE
     else:
-        mode = mode | gtk.gdkgl.MODE_SINGLE
+        mode = mode | MODE_SINGLE
     return mode
 
+FRIENDLY_MODE_NAMES = {
+                       MODE_RGB         : "RGB",
+                       MODE_RGBA        : "RGBA",
+                       MODE_ALPHA       : "ALPHA",
+                       MODE_DEPTH       : "DEPTH",
+                       MODE_DOUBLE      : "DOUBLE",
+                       MODE_SINGLE      : "SINGLE",
+                       }
+
 def get_MODE_names(mode):
-    import gtk.gdkgl
-    friendly_mode_names = {gtk.gdkgl.MODE_RGB       : "RGB",
-                           gtk.gdkgl.MODE_RGBA      : "RGBA",
-                           gtk.gdkgl.MODE_ALPHA     : "ALPHA",
-                           gtk.gdkgl.MODE_DEPTH     : "DEPTH",
-                           gtk.gdkgl.MODE_DOUBLE    : "DOUBLE",
-                           gtk.gdkgl.MODE_SINGLE    : "SINGLE"}
-    friendly_modes = [v for k,v in friendly_mode_names.items() if k>0 and (k&mode)==k]
+    global FRIENDLY_MODE_NAMES
+    friendly_modes = [v for k,v in FRIENDLY_MODE_NAMES.items() if k>0 and (k&mode)==k]
     #special case for single (value is zero!)
-    if not (mode&gtk.gdkgl.MODE_DOUBLE==gtk.gdkgl.MODE_DOUBLE):
+    if not (mode&MODE_DOUBLE==MODE_DOUBLE):
         friendly_modes.append("SINGLE")
     return friendly_modes
 
@@ -98,10 +129,11 @@ def raise_error(msg):
 gl_check_error = raise_error
 
 
+_version_warning_shown = False
 #support for memory views requires Python 2.7 and PyOpenGL 3.1
-def is_pyopengl_memoryview_safe(pyopengl_version):
-    if sys.version_info[:2]<(2,7):
-        #memoryview type requires python 2.7
+def is_pyopengl_memoryview_safe(pyopengl_version, accel_version):
+    if accel_version is not None and pyopengl_version!=accel_version:
+        #mismatch is not safe!
         return False
     vsplit = pyopengl_version.split('.')
     if vsplit[:2]<['3','1']:
@@ -140,10 +172,14 @@ def check_functions(*functions):
     else:
         log("All the required OpenGL functions are available: %s " % (", ".join(available)))
 
+
 #sanity checks: OpenGL version and fragment program support:
-def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=False):
-    if not gldrawable.gl_begin(glcontext):
-        raise ImportError("gl_begin failed on %s" % gldrawable)
+def check_GL_support(widget, force_enable=False):
+    from xpra.client.gl.gtk_compat import GLContextManager
+    with GLContextManager(widget):
+        return do_check_GL_support(force_enable)
+
+def do_check_GL_support(force_enable):
     props = {}
     try:
         #log redirection:
@@ -164,7 +200,7 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
         import OpenGL
         props["pyopengl"] = OpenGL.__version__
         from OpenGL.GL import GL_VERSION, GL_EXTENSIONS
-        from OpenGL.GL import glGetString, glGetInteger
+        from OpenGL.GL import glGetString, glGetInteger, glGetIntegerv
         gl_version_str = glGetString(GL_VERSION)
         if gl_version_str is None:
             gl_check_error("OpenGL version is missing - cannot continue")
@@ -179,16 +215,35 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
         else:
             log("found valid OpenGL version: %s.%s", gl_major, gl_minor)
 
-	from OpenGL import version as OpenGL_version
+        from OpenGL import version as OpenGL_version
+        pyopengl_version = OpenGL_version.__version__
         try:
-            import OpenGL_accelerate
+            import OpenGL_accelerate            #@UnresolvedImport
+            accel_version = OpenGL_accelerate.__version__
+            props["accelerate"] = accel_version
+            log("OpenGL_accelerate version %s", accel_version)
         except:
+            log("OpenGL_accelerate not found")
             OpenGL_accelerate = None
-        props["zerocopy"] = is_pyopengl_memoryview_safe(OpenGL_version.__version__) and bool(OpenGL_accelerate)
+            accel_version = None
+
+        if accel_version is not None and pyopengl_version!=accel_version:
+            global _version_warning_shown
+            if not _version_warning_shown:
+                log.warn("Warning: version mismatch between PyOpenGL and PyOpenGL-accelerate")
+                log.warn(" this may cause crashes")
+                _version_warning_shown = True
+        vsplit = pyopengl_version.split('.')
+        #we now require PyOpenGL 3.1 or later
+        if vsplit[:3]<['3','1'] and not force_enable:
+            gl_check_error("PyOpenGL version %s is too old and buggy" % pyopengl_version)
+            return {}
+        props["zerocopy"] = bool(OpenGL_accelerate) and is_pyopengl_memoryview_safe(pyopengl_version, accel_version)
 
         try:
-            extensions = glGetString(GL_EXTENSIONS).split(" ")
+            extensions = glGetString(GL_EXTENSIONS).decode().split(" ")
         except:
+            log("error querying extensions", exc_info=True)
             extensions = []
             gl_check_error("OpenGL could not find the list of GL extensions - does the graphics driver support OpenGL?")
         log("OpenGL extensions found: %s", ", ".join(extensions))
@@ -201,11 +256,17 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
             pass
 
         from OpenGL.GL import GL_RENDERER, GL_VENDOR, GL_SHADING_LANGUAGE_VERSION
+        def fixstring(v):
+            try:
+                return str(v).strip()
+            except:
+                return str(v)
         for d,s,fatal in (("vendor",     GL_VENDOR,      True),
                           ("renderer",   GL_RENDERER,    True),
-                          ("shading language version", GL_SHADING_LANGUAGE_VERSION, False)):
+                          ("shading-language-version", GL_SHADING_LANGUAGE_VERSION, False)):
             try:
                 v = glGetString(s)
+                v = fixstring(v.decode())
                 log("%s: %s", d, v)
             except:
                 if fatal:
@@ -214,25 +275,36 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
                     log("OpenGL property '%s' is missing", d)
                 v = ""
             props[d] = v
+        vendor = props["vendor"]
+        version_req = VERSION_REQ.get(vendor)
+        if version_req:
+            req_maj, req_min = version_req
+            if gl_major<req_maj or (gl_major==req_maj and gl_minor<req_min):
+                if force_enable:
+                    log.warn("Warning: '%s' OpenGL driver requires version %i.%i", vendor, req_maj, req_min)
+                    log.warn(" version %i.%i was found", gl_major, gl_minor)
+                else:
+                    gl_check_error("OpenGL version %i.%i is too old, %i.%i is required for %s" % (gl_major, gl_minor, req_maj, req_min, vendor))
 
         from OpenGL.GLU import gluGetString, GLU_VERSION, GLU_EXTENSIONS
-        for d,s in {"GLU version": GLU_VERSION, "GLU extensions":GLU_EXTENSIONS}.items():
+        for d,s in {"GLU.version": GLU_VERSION, "GLU.extensions":GLU_EXTENSIONS}.items():
             v = gluGetString(s)
+            v = v.decode()
             log("%s: %s", d, v)
             props[d] = v
 
-        blacklisted = None
-        whitelisted = None
-        for k,vlist in BLACKLIST.items():
-            v = props.get(k)
-            if v in vlist:
-                log("%s '%s' found in blacklist: %s", k, v, vlist)
-                blacklisted = k, v
-        for k,vlist in WHITELIST.items():
-            v = props.get(k)
-            if v in vlist:
-                log("%s '%s' found in whitelist: %s", k, v, vlist)
-                whitelisted = k, v
+        def match_list(thelist, listname):
+            for k,vlist in thelist.items():
+                v = props.get(k)
+                matches = [x for x in vlist if v.find(x)>=0]
+                if matches:
+                    log("%s '%s' found in %s: %s", k, v, listname, vlist)
+                    return (k, v)
+                log("%s '%s' not found in %s: %s", k, v, listname, vlist)
+            return None
+        blacklisted = match_list(BLACKLIST, "blacklist")
+        greylisted = match_list(GREYLIST, "greylist")
+        whitelisted = match_list(WHITELIST, "whitelist")
         if blacklisted:
             if whitelisted:
                 log.info("%s '%s' enabled (found in both blacklist and whitelist)", *whitelisted)
@@ -240,13 +312,21 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
                 log.warn("Warning: %s '%s' is blacklisted!", *blacklisted)
             else:
                 gl_check_error("%s '%s' is blacklisted!" % (blacklisted))
+        safe = bool(whitelisted) or not bool(blacklisted)
+        if safe and sys.version_info[0]>2:
+            log.warn("Warning: OpenGL python3 support is not enabled by default")
+            safe = False
+        if safe and greylisted:
+            log.warn("Warning: %s '%s' is greylisted,", *greylisted)
+            log.warn(" you may want to turn off OpenGL if you encounter bugs")
+        props["safe"] = safe
 
         #check for specific functions we need:
         from OpenGL.GL import glActiveTexture, glTexSubImage2D, glTexCoord2i, \
             glViewport, glMatrixMode, glLoadIdentity, glOrtho, \
             glEnableClientState, glGenTextures, glDisable, \
             glBindTexture, glPixelStorei, glEnable, glBegin, glFlush, \
-            glTexParameteri, \
+            glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple, \
             glTexImage2D, \
             glMultiTexCoord2i, \
             glVertex2i, glEnd
@@ -254,7 +334,7 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
             glViewport, glMatrixMode, glLoadIdentity, glOrtho, \
             glEnableClientState, glGenTextures, glDisable, \
             glBindTexture, glPixelStorei, glEnable, glBegin, glFlush, \
-            glTexParameteri, \
+            glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple, \
             glTexImage2D, \
             glMultiTexCoord2i, \
             glVertex2i, glEnd)
@@ -268,6 +348,7 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
             log.warn("OpenGL glEnablei is not available, disabling transparency")
             global GL_ALPHA_SUPPORTED
             GL_ALPHA_SUPPORTED = False
+        props["transparency"] = GL_ALPHA_SUPPORTED
 
         #check for framebuffer functions we need:
         from OpenGL.GL.ARB.framebuffer_object import GL_FRAMEBUFFER, \
@@ -307,20 +388,30 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
             try:
                 from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE
                 rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
-            except ImportError, e:
+            except ImportError as e:
                 log("OpenGL: %s", e)
                 log("using GL_MAX_TEXTURE_SIZE=%s as default", texture_size)
-        except Exception, e:
+        except Exception as e:
             emsg = str(e)
             if hasattr(e, "description"):
                 emsg = e.description
             gl_check_error("unable to query max texture size: %s" % emsg)
             return props
 
-        if min_texture_size>texture_size or min_texture_size>rect_texture_size:
-            gl_check_error("The texture size is too small: %s" % texture_size)
-        else:
-            log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB=%s, GL_MAX_TEXTURE_SIZE=%s", rect_texture_size, texture_size)
+        log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s, GL_MAX_TEXTURE_SIZE=%s", rect_texture_size, texture_size)
+        texture_size_limit = min(rect_texture_size, texture_size)
+        props["texture-size-limit"] = texture_size_limit
+
+        try:
+            from OpenGL.GL import GL_MAX_VIEWPORT_DIMS
+            v = glGetIntegerv(GL_MAX_VIEWPORT_DIMS)
+            max_viewport_dims = v[0], v[1]
+            assert max_viewport_dims[0]>=texture_size_limit and max_viewport_dims[1]>=texture_size_limit
+            log("GL_MAX_VIEWPORT_DIMS=%s", max_viewport_dims)
+        except ImportError as e:
+            log.error("Error querying max viewport dims: %s", e)
+            max_viewport_dims = texture_size_limit, texture_size_limit
+        props["max-viewport-dims"] = max_viewport_dims
         return props
     finally:
         for x in alogger.handlers[0].records:
@@ -328,9 +419,10 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
             msg = x.getMessage().replace("No OpenGL_accelerate module loaded: ", "")
             if msg=="No module named OpenGL_accelerate":
                 msg = "missing accelerate module"
-            if msg!="OpenGL_accelerate module loaded":
-                msg = "PyOpenGL warning: %s" % msg
-            log.info(msg)
+            if msg=="OpenGL_accelerate module loaded":
+                log.info(msg)
+            else:
+                log.warn("PyOpenGL warning: %s", msg)
 
         #format handler messages:
         STRIP_LOG_MESSAGE = "Unable to load registered array format handler "
@@ -367,7 +459,10 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
                 m = m.replace("accelerators", "").replace("accelerator", "").strip()
                 missing_accelerators.append(m)
                 continue
-            log.info(msg)
+            elif msg.startswith("Using accelerated"):
+                log(msg)
+            else:
+                log.info(msg)
         if missing_accelerators:
             log.info("OpenGL accelerate missing: %s", ", ".join(missing_accelerators))
 
@@ -379,39 +474,69 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
         restore_logger(alogger)
         restore_logger(arlogger)
         restore_logger(clogger)
-        gldrawable.gl_end()
 
-def check_support(min_texture_size=0, force_enable=False, check_colormap=False):
+
+def check_support(force_enable=False, check_colormap=False):
     #platform checks:
     from xpra.platform.gui import gl_check
     warning = gl_check()
     if warning:
-        gl_check_error(warning)
+        if force_enable:
+            log.warn("Warning: trying to continue despite '%s'" % warning)
+        else:
+            gl_check_error(warning)
 
     props = {}
-    import gtk.gdk
-    import gtk.gdkgl, gtk.gtkgl
-    assert gtk.gdkgl is not None and gtk.gtkgl is not None
-    log("pygdkglext version=%s", gtk.gdkgl.pygdkglext_version)
-    props["pygdkglext_version"] = gtk.gdkgl.pygdkglext_version
-    log("gtkglext_version=%s", gtk.gdkgl.gdkglext_version)
-    props["gdkglext_version"] = gtk.gdkgl.gdkglext_version
-    log("pygdkglext OpenGL version=%s", gtk.gdkgl.query_version())
-    props["gdkgl_version"] = gtk.gdkgl.query_version()
-    display_mode = get_DISPLAY_MODE()
+    #this will import gtk.gtkgl / gdkgl or gi.repository.GtkGLExt / GdkGLExt:
     try:
-        glconfig = gtk.gdkgl.Config(mode=display_mode)
-    except gtk.gdkgl.NoMatches, e:
-        log("no match: %s, toggling double-buffering", e)
-        display_mode &= ~gtk.gdkgl.MODE_DOUBLE
-        glconfig = gtk.gdkgl.Config(mode=display_mode)
+        from xpra.client.gl.gtk_compat import get_info, gdkgl, Config_new_by_mode, GLDrawingArea
+    except RuntimeError as e:
+        gl_check_error(str(e))
+        return {}
+    props.update(get_info())
+    display_mode = get_DISPLAY_MODE()
+    glconfig = Config_new_by_mode(display_mode)
+    if glconfig is None and CAN_DOUBLE_BUFFER:
+        log("trying to toggle double-buffering")
+        display_mode &= ~MODE_DOUBLE
+        glconfig = Config_new_by_mode(display_mode)
+    if not glconfig:
+        gl_check_error("cannot setup an OpenGL context")
     props["display_mode"] = get_MODE_names(display_mode)
-    props["glconfig"] = glconfig
-    props["has_alpha"] = glconfig.has_alpha()
-    props["rgba"] = glconfig.is_rgba()
+    #on OSX, we had to patch out get_depth...
+    #so take extra precautions when querying properties:
+    for x,fn_name in {
+        "has_alpha"           : "has_alpha",
+        "rgba"                : "is_rgba",
+        "stereo"              : "is_stereo",
+        "double-buffered"     : "is_double_buffered",
+        "depth"               : "get_depth",
+        "has-depth-buffer"    : "has_depth_buffer",
+        "has-stencil-buffer"  : "has_stencil_buffer",
+        }.items():
+        fn = getattr(glconfig, fn_name, None)
+        if fn:
+            props[x] = fn()
+        else:
+            log("%s does not support %s()", glconfig, fn_name)
+    for x in ("RED_SIZE", "GREEN_SIZE", "BLUE_SIZE", "ALPHA_SIZE",
+              "AUX_BUFFERS", "DEPTH_SIZE", "STENCIL_SIZE",
+              "ACCUM_RED_SIZE", "ACCUM_GREEN_SIZE", "ACCUM_BLUE_SIZE",
+              "SAMPLE_BUFFERS", "SAMPLES"):
+        prop = getattr(gdkgl, x)
+        if not prop:
+            continue
+        try:
+            v = glconfig.get_attrib(prop)[0]
+            props[x.lower().replace("_", "-")] = v
+        except:
+            pass
     log("GL props=%s", props)
-    assert gtk.gdkgl.query_extension()
-    glcontext, gldrawable, glext, w = None, None, None, None
+
+    from xpra.gtk_common.gtk_util import import_gtk, gdk_window_process_all_updates
+    gtk = import_gtk()
+    assert gdkgl.query_extension()
+    glext, w = None, None
     try:
         #ugly code for win32 and others (virtualbox broken GL drivers)
         #for getting a GL drawable and context: we must use a window...
@@ -424,7 +549,7 @@ def check_support(min_texture_size=0, force_enable=False, check_colormap=False):
         w = gtk.Window()
         w.set_decorated(False)
         vbox = gtk.VBox()
-        glarea = gtk.gtkgl.DrawingArea(glconfig)
+        glarea = GLDrawingArea(glconfig)
         glarea.set_size_request(32, 32)
         vbox.add(glarea)
         vbox.show_all()
@@ -432,39 +557,43 @@ def check_support(min_texture_size=0, force_enable=False, check_colormap=False):
         #we don't need to actually show the window!
         #w.show_all()
         glarea.realize()
-        gtk.gdk.window_process_all_updates()
-        gldrawable = glarea.get_gl_drawable()
-        glcontext = glarea.get_gl_context()
+        gdk_window_process_all_updates()
 
-        gl_props = check_GL_support(gldrawable, glcontext, min_texture_size, force_enable)
+        gl_props = check_GL_support(glarea, force_enable)
 
         if check_colormap:
-            rgb_visual = w.get_screen().get_rgb_visual()
-            rgba_visual = w.get_screen().get_rgba_visual()
-            gl_props["rgb_visual"] = visual_to_str(rgb_visual)
-            gl_props["rgba_visual"] = visual_to_str(rgba_visual)
+            s = w.get_screen()
+            for x in ("rgb_visual", "rgba_visual", "system_visual"):
+                try:
+                    visual = getattr(s, "get_%s" % x)()
+                    gl_props[x] = visual_to_str(visual)
+                except:
+                    pass
             #i = 0
-            #for v in w.get_screen().list_visuals():
+            #for v in s.list_visuals():
             #    gl_props["visual[%s]" % i] = visual_to_str(v)
             #    i += 1
     finally:
         if w:
             w.destroy()
-        del glcontext, gldrawable, glext, glconfig
+        del glext, glconfig
     props.update(gl_props)
     return props
 
 
-def main():
-    from xpra.platform import init,clean
+def main(force_enable=False):
+    from xpra.platform import program_context
     from xpra.platform.gui import init as gui_init
-    from xpra.util import pver
-    try:
-        init("OpenGL-Check")
+    from xpra.util import print_nested_dict
+    from xpra.log import enable_color
+    with program_context("OpenGL-Check"):
         gui_init()
+        enable_color()
         verbose = "-v" in sys.argv or "--verbose" in sys.argv
         if verbose:
             log.enable_debug()
+            from xpra.client.gl.gtk_compat import log as clog
+            clog.enable_debug()
         #replace ImportError with a log message:
         global gl_check_error
         errors = []
@@ -472,7 +601,7 @@ def main():
             log.error("ERROR: %s", msg)
             errors.append(msg)
         gl_check_error = log_error
-        props = check_support(0, True, verbose)
+        props = check_support(force_enable, verbose)
         log.info("")
         if len(errors)>0:
             log.info("OpenGL errors:")
@@ -480,14 +609,8 @@ def main():
                 log.info("  %s", e)
         log.info("")
         log.info("OpenGL properties:")
-        for k in sorted(props.keys()):
-            v = props[k]
-            #skip not human readable:
-            if k not in ("extensions", "glconfig"):
-                log.info("* %s : %s", str(k).ljust(24), pver(v))
+        print_nested_dict(props)
         return len(errors)
-    finally:
-        clean()
 
 
 if __name__ == "__main__":
